open Format
open Lexing
open Trs_ast

let verbose = ref false

let cime_output = ref false

let extension f =
  try
    let i = String.rindex f '.' in
    String.sub f (i+1) (String.length f - i - 1)
  with Not_found -> ""

let rm_extension f =
  try
    let i = String.rindex f '.' in
    String.sub f 0 i 
  with Not_found -> f

(* db output *)

let db_output = ref false

let db_ch = ref stdout

let db_fmt = ref std_formatter

let db_trs_pbs = ref []

let db_srs_pbs = ref []

let db_lp_pbs = ref []

let db_pb pbs f =
  let dir = Filename.basename (Filename.dirname f) in
  let file = Filename.basename f in
  let name = rm_extension file in
  fprintf !db_fmt "%s.%s = {@." dir file;
  pbs := (dir ^ "." ^ file) :: !pbs;
  fprintf !db_fmt "name  = \"%s - %s\";@." dir name;
  fprintf !db_fmt "file  = \"%s\";@." f

let get_categories ast =
  List.fold_left
    (fun (cond,t,i,cs) decl ->
       match decl with
	 | RulesDecl l -> 
	     let cond =
	       List.fold_left
		 (fun cond (c,_,_) -> 
		    match c with [] -> cond | _ -> true)
		 cond l
	     in (cond,t,i,cs)
	 | TheoryDecl _ -> (cond,true,i,cs)
	 | StrategyDecl Innermost -> (cond,t,true,cs)
	 | StrategyDecl (ContextSensitive _) -> (cond,t,i,true)
	 |_ -> (cond,t,i,cs))
    (false,false,false,false)
    ast

let check_has f b1 b2 msg =
  if b1 && b2 then
    printf "Warning: problem %s has both %s.@." f msg 

let db_trs f ast = 
  db_pb db_trs_pbs f;
  let (has_cond,has_theory,has_innermost,has_context_sensitive) =
    get_categories ast
  in
  check_has f has_cond has_theory "conditions and theory";
  check_has f has_cond has_innermost "conditions and innermost strategy";
  check_has f has_cond has_context_sensitive "conditions and CS strategy";
  check_has f has_theory has_innermost "theory and innermost strategy";
  check_has f has_theory has_context_sensitive "theory and CS strategy";
  check_has f has_innermost has_context_sensitive "innermost and CS strategy";
  fprintf !db_fmt "conditional = %b;@." has_cond;
  fprintf !db_fmt "theory  = %b;@." has_theory;
  fprintf !db_fmt "innermost  = %b;@." has_innermost;
  fprintf !db_fmt "contextsensitive = %b;@." has_context_sensitive;
  fprintf !db_fmt "}@.@."

let db_srs f ast = 
  db_pb db_srs_pbs f;
  fprintf !db_fmt "}@.@."

let get_query f ast = 
  match
    List.fold_left
      (fun query decl ->
	 match decl with
	   | Lp_ast.Clause _ -> query
	   | Lp_ast.Query q -> 
	       match query with
		 | None -> Some q
		 | Some _ -> 
		     printf "Warning: too many queries in %s@." f; 
		     query)
      None ast
  with
    | None -> 
	printf "Warning: no query in %s@." f;
	"no_query_given()"
    | Some q -> 
	match q with
	  | Lp_ast.Pred((_,id),l) ->
	      try
		let args,_ =
		  List.fold_right
		    (fun a (args,sep) ->
		       match a with
			 | Lp_ast.Term((_,"i"),[]) -> ("i"^sep^args,",")
			 | Lp_ast.Term((_,"o"),[]) -> ("o"^sep^args,",")
			 | _ -> raise Exit)
		    l ("",")")
		in
		id ^ "(" ^ args
	      with
		  Exit -> 
		    printf "Warning: ill-formed query in %s@." f; 
		    "ill_formed_query()"



let db_lp f ast = 
  db_pb db_lp_pbs f;
  let query = get_query f ast in
  fprintf !db_fmt "query = \"%s\";@." query;
  fprintf !db_fmt "}@.@."

let db_list name l =
  match l with
    | [] -> fprintf !db_fmt "%s = [ ];@." name;
    | p::r ->
	  fprintf !db_fmt "@[%s = [ %s" name p;
	  List.iter (fun f -> fprintf !db_fmt ";@ %s" f) r;
	  fprintf !db_fmt "];@]@."

let db_main () =
  fprintf !db_fmt "_main = {@.";
  fprintf !db_fmt "trs_tools = [ aprove ; cime ; matchbox ; ttt] ;@.";
  db_list "trs_pbs" !db_trs_pbs;
  fprintf !db_fmt "srs_tools = [ cime ; matchbox ; torpa] ;@.";
  db_list "srs_pbs" !db_srs_pbs;
  fprintf !db_fmt "lp_tools = [ aprove ; talp] ;@.";
  db_list "lp_pbs" !db_lp_pbs;
  fprintf !db_fmt "}@.";
  close_out !db_ch





(* main *)

let srs_files = ref []

let trs_files = ref []

let lp_files = ref []

let rec parse_args args =
  match args with
    | [] -> ()
    | "-v"::rem -> verbose := true; parse_args rem
    | "-cime"::rem -> cime_output := true; parse_args rem
    | "-db"::f::rem -> 
	db_output := true; 
	db_ch := open_out f; 	
	db_fmt := formatter_of_out_channel !db_ch;
	parse_args rem
    | "-trs"::f::rem -> 
	trs_files := f :: !trs_files; parse_args rem
    | "-srs"::f::rem -> 
	srs_files := f :: !srs_files; parse_args rem
    | f::rem ->
	begin
	  match extension f with
	    | "trs" ->
		trs_files := f :: !trs_files; parse_args rem
	    | "srs" ->
		srs_files := f :: !srs_files; parse_args rem
	    | "pl" ->
		lp_files := f :: !lp_files; parse_args rem
	    | _ ->
		eprintf "don't known what to do with argument %s@." f;
		exit 2;
	end
	

let check_trs f =
  let c = open_in f in
  let lb = from_channel c in
  try
    lb.lex_curr_p <-
    { pos_fname = f ; pos_lnum = 1; pos_cnum = 0; pos_bol = 0};
    let ast = Trs_parser.spec Trs_lexer.token lb in
    if !verbose then printf "parse ok@.";
    close_in c;
    let env = Trs_sem.spec ast in
    if !verbose then 
      begin
	printf "semantics ok, function symbols are:@.";
	List.iter
	  (fun (id,arity) ->
	     match arity with
	       | Trs_sem.Var -> ()
	       | Trs_sem.Function x -> printf "%s: %d@." id !x)
	  env
      end;
    if !cime_output then Cime.output f env ast;
    if !db_output then db_trs f ast
  with 
    | Parsing.Parse_error ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: syntax error.@." Trs_ast.print_loc loc;
	close_in c;
	exit 1
    | Trs_lexer.Lexing_error msg ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: lexer error, %s.@." Trs_ast.print_loc loc msg;
	close_in c;
	exit 1
    | Trs_sem.Sem_error(loc,msg) ->
	eprintf "%a: %s.@." Trs_ast.print_loc loc msg;
	exit 1

let check_srs f =
  let c = open_in f in
  let lb = from_channel c in
  try
    lb.lex_curr_p <-
    { pos_fname = f ; pos_lnum = 1; pos_cnum = 0; pos_bol = 0};
    let ast = Srs_parser.spec Srs_lexer.token lb in
    if !verbose then printf "parse ok@.";
    close_in c;
    if !cime_output then Cime.output_srs f ast;
    if !db_output then db_srs f ast
  with 
    | Parsing.Parse_error ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: syntax error.@." Trs_ast.print_loc loc;
	close_in c;
	exit 1

let check_lp f =
  let c = open_in f in
  let lb = from_channel c in
  try
    lb.lex_curr_p <-
    { pos_fname = f ; pos_lnum = 1; pos_cnum = 0; pos_bol = 0};
    let ast = Lp_parser.spec Lp_lexer.token lb in
    if !verbose then printf "parse ok@.";
    close_in c;
    if !db_output then db_lp f ast
  with 
    | Parsing.Parse_error ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: syntax error.@." Trs_ast.print_loc loc;
	close_in c;
	exit 1
    | Failure msg ->
	let loc = Lexing.lexeme_start_p lb in
	eprintf "%a: Failure: %s.@." Trs_ast.print_loc loc msg;
	exit 1
  
let main () =
  try
    parse_args (List.tl (Array.to_list Sys.argv));
    List.iter check_trs !trs_files;
    List.iter check_srs !srs_files;
    List.iter check_lp !lp_files;
    if !db_output then db_main ();
    exit 0
      
  with
    | Sys_error msg ->
	eprintf "System error: %s.@." msg;
	exit 1
    | Failure msg ->
	eprintf "Failure: %s.@." msg;
	exit 1

let _ = Printexc.catch main ()


	     
